var definitions = require("./lib/definitions");
var LOG = require("./lib/log");

module.exports = function(obj) {
    var t = obj.types;

    var RUNTIME_MODULE_NAME = "check-is-es5";

    var HELPER_BLACKLIST = ["interopRequireWildcard", "interopRequireDefault"];

    var es6FeatureApiHash = {};

    buildHash(definitions.methods, es6FeatureApiHash);

    return {
        pre: function (file) {
            // default log level
            LOG.setLevel(1);
            // console.log(file.code);
            LOG.setSourceCode(file.code);
        },
        visitor: {
            // static reference check
            ReferencedIdentifier: function(path, state) {
                var node = path.node;
                var parent = path.node;
                var scope = path.scope;

                if (node.name === "regeneratorRuntime") {
                    LOG.error("You may use `regenerator` in your code.", path.hub.file.opts.filename, node.loc);
                    return;
                }

                if (t.isMemberExpression(parent)) return;
                if (!has(definitions.builtins, node.name)) return;
                if (scope.getBindingIdentifier(node.name)) return;

                LOG.setLevel(state.opts.log);
                LOG.error("You may use `"+ node.name +"` in your code.", path.hub.file.opts.filename, node.loc);
            },

            // arr[Symbol.iterator]() check
            CallExpression: function(path, state) {
                if (path.node.arguments.length) return;

                var callee = path.node.callee;
                if (!t.isMemberExpression(callee)) return;
                if (!callee.computed) return;
                if (!path.get("callee.property").matchesPattern("Symbol.iterator")) return;

                LOG.setLevel(state.opts.log);                
                LOG.error('', path.hub.file.opts.filename, path.node.loc);
            },

            // Symbol.iterator in arr check
            BinaryExpression: function(path, state) {
                // console.log('BinaryExpression: ', path.node);

                if (path.node.operator !== "in") return;
                if (!path.get("left").matchesPattern("Symbol.iterator")) return;

                LOG.setLevel(state.opts.log);
                LOG.error('You may traversal `Symbol` in es5!', path.hub.file.opts.filename, path.node.loc);
            },

            // member function or properties call check
            MemberExpression: {
                enter: function(path, state) {
                    if (!path.isReferenced()) return;

                    var node = path.node;
                    var obj = node.object;
                    var prop = node.property;
                    var filename = path.hub.file.opts.filename;

                    if (!t.isReferenced(obj, node)) return;
                    if (node.computed) return;

                    var name = "";
                    var objName = getName(obj);

                    LOG.setLevel(state.opts.log);

                    if (!has(es6FeatureApiHash, prop.name)) return;

                    if (!has(definitions.methods, objName)) {
                        try {
                            name = findUpScopeUntilRoot(path, objName);                        
                        } catch(e) {
                            LOG[e.type] && LOG[e.type](e.message, filename, e.loc);
                        }
                    } else {
                        name = objName;
                    }

                    var methods = get(definitions.methods, name);
                    if (!has(methods, prop.name) || prop.name === 'prototype') return;

                    // doesn't reference the global
                    if (path.scope.getBindingIdentifier(name)) return;

                    // special case Object.defineProperty to not use core-js when using string keys
                    if (obj.name === "Object" && prop.name === "defineProperty" && path.parentPath.isCallExpression()) {
                        var call = path.parentPath.node;
                        if (call.arguments.length === 3 && t.isLiteral(call.arguments[1])) return;
                    }

                    LOG.error('Your code called `'+ [name, prop.name].join('.') +'` that es5 is not support!', filename, obj.loc);
                },

                exit: function(path, state) {
                    if (state.opts.polyfill === false) return;
                    if (!path.isReferenced()) return;

                    var node = path.node;
                    var obj = node.object;
                    var name = getName(obj);
                    var filename = path.hub.file.opts.filename;
                    if (!has(definitions.builtins, name)) return;
                    if (path.scope.getBindingIdentifier(name)) return;

                    LOG.setLevel(state.opts.log);                    
                    LOG.error('Your code use `'+ name +'` that es5 is not support!', filename, obj.loc);
                }
            }
        }
    };

    function has(obj, key) {
        if (!obj || !key) {
            return false;
        }
        if (~key.indexOf('.')) {
            var keys = key.split('.');
            key = keys.shift();
            return has(obj[key], keys.join('.'));
        }
        return Object.prototype.hasOwnProperty.call(obj, key);
    }

    function get(obj, key) {
        if (!obj) {
            return undefined;
        }
        if (~key.indexOf('.')) {
            var keys = key.split('.');
            key = keys.shift();
            return get(obj[key], keys.join('.'));
        }
        return obj[key];
    }

    function getName(obj) {
        switch(obj.type) {
            case 'NewExpression':
                return getName(obj.callee);
            case 'MemberExpression':
                return getName(obj.object)+ '.' +obj.property.name;
            case 'Identifier':
                return obj.name;
        }
    }

    function buildHash(obj, hash) {
        if (!hash) return;
        var keys = Object.keys(obj);
        for (var i = 0, l = keys.length; i < l; i++) {
            var o = obj[keys[i]];
            if (typeof(o) === "object") {
                buildHash(o, hash);
            } else {
                hash[keys[i]] = true;                
            }
        }
    }

    function _throw(type, message, loc) {
        throw {
            type: type,
            message: message,
            loc: loc
        }
    }

    function findUpScopeUntilRoot(path, objName) {
        var scope = path.scope;
        var node = path.node;
        var obj = node.object;
        var prop = node.property;
        var name = "";
        // find this var bindings in scope until the root scope
        while (scope && !has(scope.bindings, objName)) scope = scope.parent;
        // can not find until root scope
        if (scope) {
            switch (scope.bindings[obj.name].kind) {
                case 'param':
                    _throw("warning", "Your code called `"+ prop.name +"`(es5 not support), with unknown param type. Make sure this param can call this api `"+ prop.name +"`!", scope.bindings[obj.name].identifier.loc);
                case 'var': 
                    if (scope.bindings[obj.name].constantViolations.length) {
                        var nowPath = scope.bindings[obj.name].constantViolations[scope.bindings[obj.name].constantViolations.length - 1];
                        var nowNode = nowPath.node;

                        if (nowNode.type !== 'AssignmentExpression') {
                            name = "";                            
                            break;
                        }
                        switch (nowNode.right.type) {
                            case 'ArrayExpression':
                                name = "Array.prototype";
                                break;
                            case 'NumericLiteral':
                                name = "Number.prototype";
                                break;
                            case 'BooleanLiteral':
                                name = "Boolean.prototype";
                                break;
                            case 'StringLiteral':
                                name = "String.prototype";
                                break;
                            case 'ObjectExpression':
                                name = "Object.prototype";
                                break;
                            case 'FunctionExpression':
                                _throw("warning", "Your code called `"+ prop.name +"`(may es5 not support), with unknown param type. Make sure this param can call this api `"+ prop.name +"`!", nowNode.loc);                            
                            case 'NewExpression':
                            case 'CallExpression':
                                name = [nowNode.right.callee.name, 'prototype'].join('.');
                                if (!has(definitions.methods, name)) {
                                    _throw("warning", "Your code called `"+ prop.name +"`(may es5 not support), with unknown param type. Make sure this param can call this api `"+ prop.name +"`!", nowNode.loc);
                                }
                                break;
                            case 'NullLiteral':
                                _throw("error", nowNode.left.name +" called `"+ prop.name +"`, but it is null!", nowNode.loc);
                                break;
                            case 'Identifier':
                                // NaN,
                                if (nowNode.left.name === 'NaN') {
                                    _throw("error", nowNode.left.name +" called `"+ prop.name +"`, but it is undefined!", nowNode.loc);
                                }
                                name = findUpScopeUntilRoot(nowPath, getName(nowNode));
                                break;
                            default:
                                name = "";
                        }
                    }
                    break;
                default:
                    _throw("error", "Unknown scope kind ["+ scope.kind +"]", scope.bindings[obj.name].loc);
            }
        } else {
            _throw("warning", "Your code called `"+ prop.name +"`(may es5 not support), with unknown caller type. Make sure this caller support this api `"+ prop.name +"`!", node.loc);
        }

        return name;
    }
}

module.exports.definitions = definitions;